from typing import Dict, Set

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand


class Text(PrivateOpenScadCommand):
    """
    Widget for creating texts. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Text.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 text: str,
                 size: float = 10.0,
                 font: str | None = None,
                 halign: str = 'left',
                 valign: str = 'baseline',
                 spacing: float = 1.0,
                 direction: str = 'ltr',
                 language: str = 'en',
                 script: str = 'latin',
                 fn: int | None = None):
        """
        Object constructor.

        :param text: The text to generate.
        :param size: The generated text has an ascent (height above the baseline) of approximately the given value.
                     Different fonts can vary somewhat and may not fill the size specified exactly, typically they
                     render slightly smaller. On a metric system, a size of 25.4 (1" imperial) will correspond to
                     100pt ⇒ a 12pt font size would be 12×0.254 for metric conversion or 0.12 in imperial.
        :param font: The name of the font that should be used. This is not the name of the font file, but the logical
                     font name (internally handled by the fontconfig library). This can also include a style parameter,
                     see below. A list of installed fonts & styles can be obtained using the font list dialog (Help ->
                     Font List).
        :param halign: The horizontal alignment for the text. Possible values are "left", "center" and "right".
        :param valign: The vertical alignment for the text. Possible values are "top", "center", "baseline" and
                       "bottom".
        :param spacing: Factor to increase/decrease the character spacing. The default value of 1 results in the normal
                        spacing for the font, giving a value greater than 1 causes the letters to be spaced further
                        apart.
        :param direction: Direction of the text flow. Possible values are "ltr" (left-to-right), "rtl" (right-to-left),
                          "ttb" (top-to-bottom) and "btt" (bottom-to-top).
        :param language: The language of the text (e.g., "en", "ar", "ch").
        :param script: The script of the text (e.g., "latin", "arabic", "hani").
        :param fn: Used for subdividing the curved path segments provided by freetype.
        """
        PrivateOpenScadCommand.__init__(self, command='text', args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'fn': '$fn'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'size'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_scales(self) -> Set[str]:
        """
        Returns the set with arguments that are scales and factors.
        """
        return {'spacing'}

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def text(self) -> str:
        """
        Returns the text to generate.
        """
        return self._args.get('text')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> int:
        """
        Returns the font size.
        """
        return self._args.get('size')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def font(self) -> str | None:
        """
        Returns the name of the font.
        """
        return self._args.get('font')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def halign(self) -> str:
        """
        Returns the horizontal alignment for the text.
        """
        return self._args.get('halign')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def valign(self) -> str:
        """
        Returns the vertical alignment for the text.
        """
        return self._args.get('valign')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def spacing(self) -> float:
        """
        Returns the factor to increase/decrease the character spacing.
        """
        return self._args.get('spacing')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def direction(self) -> str:
        """
        Returns the direction of the text flow.
        """
        return self._args.get('direction')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def language(self) -> str:
        """
        Returns the language of the text.
        """
        return self._args.get('language')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def script(self) -> str:
        """
        Returns the script of the text.
        """
        return self._args.get('script')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns used for subdividing the curved path segments provided by freetype.
        """
        return self._args.get('fn')

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.boolean.Union import Union
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget


class Empty(ScadWidget):
    """
    Widget for creating empty OpenSCAD nodes.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return Union(children=[])

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Set

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2
from super_scad.type.Vector3 import Vector3


class PrivateRotate(PrivateSingleChildOpenScadCommand):
    """
    Rotates its child widget about the axis of the coordinate system or around an arbitrary axis. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#rotate.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float | Vector2 | Vector3,
                 vector: Vector3 | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param angle: The angle of the rotation.
        :param vector: The vector of the rotation.
        :param child: The widget to be rotated.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='rotate', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'angle': 'a', 'vector': 'v'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return {'a'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'v'}

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateScale import PrivateScale
from super_scad.type.Vector2 import Vector2


class Scale2D(ScadSingleChildParent):
    """
    Scales its child widget using a specified scaling factor.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 factor: Vector2 | float | None = None,
                 factor_x: float | None = None,
                 factor_y: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param factor: The scaling factor along all two the axes.
        :param factor_x: The scaling factor along the x-axis.
        :param factor_y: The scaling factor along the y-axis.
        :param child: The child to be scaled.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'factor'}, {'factor_x', 'factor_y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def factor(self) -> Vector2:
        """
        Returns the scaling factor along all two axes.
        """
        return Vector2(self.factor_x, self.factor_y)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def factor_x(self) -> float:
        """
        Returns the scaling factor along the x-axis.
        """
        if 'factor' in self._args:
            if isinstance(self._args['factor'], float):
                return self._args['factor']

            return self._args['factor'].x

        return self._args.get('factor_x', 1.0)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def factor_y(self) -> float:
        """
        Returns the scaling factor along the y-axis.
        """
        if 'factor' in self._args:
            if isinstance(self._args['factor'], float):
                return self._args['factor']

            return self._args['factor'].y

        return self._args.get('factor_y', 1.0)

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateScale(factor=self.factor, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



import math
import random
from abc import ABC, abstractmethod
from typing import List, Set

from super_scad.d2.helper.PolygonSideExtender import PolygonSideExtender
from super_scad.d2.private.PrivatePolygon import PrivatePolygon
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type import Vector2
from super_scad.type.Angle import Angle


# class PolygonMixin(ScadWidget, ABC):
class PolygonMixin(ABC):
    """
    A mixin for all polygonal and polygonal like widgets in SuperSCAD.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 extend_sides_by_eps: bool | List[bool] | Set[int] | None):
        """
        Object constructor.

        :param extend_sides_by_eps: Whether to extend sides by eps for a clear overlap.
        """
        self._inner_angles: List[float] | None = None
        """
        The inner angles of the polygon (in the same order as the primary points).
        """

        self._normal_angles: List[float] | None = None
        """
        The absolute angles of the normal of each node.
        """

        self._is_clockwise: bool | None = None
        """
        Whether the nodes of the polygon are in a clockwise order.
        """

        self._args['extend_sides_by_eps'] = extend_sides_by_eps

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def is_clockwise_order(nodes: List[Vector2], delta: float) -> bool:
        """
        Returns whether the nodes of a polygon are given in a clockwise order.

        @param nodes: The nodes of the polygon.
        @param delta: The minimum distance between the line segment and the nodes.
        """
        radius = 0.0
        for node in nodes:
            radius = max(radius, abs(node.x), abs(node.y))

        n = len(nodes)
        for index1 in range(10 * n):
            min_distance = radius
            for index2 in range(1, n):
                if index2 != index1 % n:
                    min_distance = min(min_distance, Vector2.distance(nodes[index1 % n], nodes[index2]))

            p1 = nodes[(index1 - 1) % n]
            p2 = nodes[index1 % n]
            p3 = nodes[(index1 + 1) % n]

            leg1 = (p2 - p1)
            leg2 = (p2 - p3)
            height = abs(Vector2.cross_product(leg1, leg2)) / Vector2.distance(leg2, leg1)
            q1 = p2 + Vector2.from_polar_coordinates(random.uniform(0.25, 0.75) * min(height, min_distance),
                                                     0.5 * (leg1.angle + leg2.angle))
            if not PolygonMixin._to_close(p1, p3, q1, delta):
                q2 = Vector2.from_polar_coordinates(2.0 * radius, random.uniform(0.0, 360.0))
                number_of_intersections = PolygonMixin._count_intersections(nodes, q1, q2, delta)
                if number_of_intersections is not None:
                    orientation = Vector2.orientation(p1, p2, q1)
                    assert orientation != 0.0

                    return (orientation > 0.0) == (number_of_intersections % 2 == 1)

        raise ValueError('Not a proper polygon.')

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _to_close(segment_start: Vector2, segment_end: Vector2, node: Vector2, delta: float) -> bool:
        """
        Returns whether a node is to close to a line segment for reliable computation of the separation between line
        segments and nodes.

        :param segment_start: The start point of the line segment.
        :param segment_end: The end point of the line segment.
        :param node: The node.
        :param delta: The minimum distance between nodes, vertices and line segments for reliable computation of the
                      separation between line segments and nodes.

        """
        tmp1 = segment_end - segment_start
        tmp2 = (node - segment_start).rotate(-tmp1.angle)

        return -delta <= tmp2.x <= (tmp1.length + delta) and -delta <= tmp2.y <= delta

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _do_intersect(segment_start: Vector2,
                      segment_end: Vector2,
                      vertex_start: Vector2,
                      vertex_end: Vector2) -> bool:
        """
        Returns whether a line segment and a vertex intersect.

        @param segment_start: The start point of the line segment.
        @param segment_end: The end point of the line segment.
        @param vertex_start: The start point of the vertex.
        @param vertex_end:  The end point of the vertex.
        """
        o1 = Vector2.orientation(segment_start, segment_end, vertex_start)
        o2 = Vector2.orientation(segment_start, segment_end, vertex_end)
        o3 = Vector2.orientation(vertex_start, vertex_end, segment_start)
        o4 = Vector2.orientation(vertex_start, vertex_end, segment_end)

        return (math.copysign(1, o1) != math.copysign(1, o2)) and (math.copysign(1, o3) != math.copysign(1, o4))

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _count_intersections(nodes: List[Vector2],
                             segment_start: Vector2,
                             segment_end: Vector2,
                             delta: float) -> int | None:
        """
        Returns the number of intersections between a line segment (segment_start, segment_end) and the vertices of the
        polygon.

        :param nodes: The nodes of the polygon.
        :param segment_start: Start point of the line segment.
        :param segment_end: End point of the line segment.
        :param delta: The minimum distance between nodes, vertices and line segments for reliable computation of the
                      separation between line segments and nodes.
        """
        intersections = 0

        n = len(nodes)
        for i in range(n):
            vertex_start = nodes[i]
            vertex_end = nodes[(i + 1) % n]

            if PolygonMixin._to_close(segment_start, segment_end, vertex_start, delta):
                return None

            if PolygonMixin._to_close(vertex_start, vertex_end, segment_start, delta):
                return None

            # segment_end is twice as far from the origin as any node from the origin.

            if PolygonMixin._do_intersect(segment_start, segment_end, vertex_start, vertex_end):
                intersections += 1

        return intersections

    # ------------------------------------------------------------------------------------------------------------------
    def _compute_angles(self, context: Context) -> None:
        """
        Returns the inner angles of the polygon (in the same order as the primary points).

        :param context: The build context.
        """
        self._inner_angles = []
        self._normal_angles = []

        nodes = self.nodes
        self._is_clockwise = self.is_clockwise_order(nodes, context.delta)

        n = len(nodes)
        for i in range(n):
            if self._is_clockwise:
                p1 = nodes[(i - 1) % n]
                p2 = nodes[i]
                p3 = nodes[(i + 1) % n]
            else:
                p1 = nodes[(i + 1) % n]
                p2 = nodes[i]
                p3 = nodes[(i - 1) % n]

            inner_angle = Angle.normalize((p3 - p2).angle - (p2 - p1).angle - 180.0)
            normal_angle = Angle.normalize((p1 - p2).angle + 0.5 * inner_angle)

            self._inner_angles.append(inner_angle)
            self._normal_angles.append(normal_angle)

    # ------------------------------------------------------------------------------------------------------------------
    def is_clockwise(self, context: Context) -> bool:
        """
        Returns whether the nodes of this polygon are in a clockwise order.

        :param context: The build context.
        """
        if self._is_clockwise is None:
            self._compute_angles(context)

        return self._is_clockwise

    # ------------------------------------------------------------------------------------------------------------------
    def inner_angles(self, context: Context) -> List[float]:
        """
        Returns the inner angles of the polygon (in the same order as the primary points).

        :param context: The build context.
        """
        if self._inner_angles is None:
            self._compute_angles(context)

        return self._inner_angles

    # ------------------------------------------------------------------------------------------------------------------
    def normal_angles(self, context: Context) -> List[float]:
        """
        Returns the absolute angles of the normal of each node.

        :param context: The build context.
        """
        if self._normal_angles is None:
            self._compute_angles(context)

        return self._normal_angles

    # ------------------------------------------------------------------------------------------------------------------
    @property
    @abstractmethod
    def nodes(self) -> List[Vector2]:
        """
        Returns the nodes of this polygon.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def extend_sides_by_eps(self) -> Set[int]:
        """
        Returns the set of sides that must be extended by eps for clear overlap.
        """
        extend_sides_by_eps = self._args.get('extend_sides_by_eps')

        if isinstance(extend_sides_by_eps, set):
            return extend_sides_by_eps

        if extend_sides_by_eps is None:
            extend_sides_by_eps = set()

        elif extend_sides_by_eps is True:
            extend_sides_by_eps = {index for index in range(self.sides)}

        elif isinstance(extend_sides_by_eps, list):
            extend_sides_by_eps = set(index for index in range(len(extend_sides_by_eps)) if extend_sides_by_eps[index])

        else:
            raise ValueError(f'Parameter extend_sides_by_eps must be a boolean, '
                             f'set of integers, a list of booleans or None, got {type(extend_sides_by_eps)}')

        self._args['extend_sides_by_eps'] = extend_sides_by_eps

        return extend_sides_by_eps

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def sides(self) -> int:
        """
        Returns the number of sides of this polygon.
        """
        return len(self.nodes)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the convexity of the polygon.
        """
        return self._args.get('convexity')

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        if self.extend_sides_by_eps:
            return self._build_polygon_extended(context)

        return self._build_polygon(context)

    # ------------------------------------------------------------------------------------------------------------------
    @abstractmethod
    def _build_polygon(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    def _build_polygon_extended(self, context: Context) -> ScadWidget:
        """
        Builds a polygon with extended sides.

        @param context: The build context.
        """
        polygon_side_extender = self._create_polygon_side_extender()
        new_nodes = polygon_side_extender.extend_sides(context=context,
                                                       nodes=self.nodes,
                                                       inner_angles=self.inner_angles(context),
                                                       normal_angles=self.normal_angles(context),
                                                       is_clockwise=self.is_clockwise(context),
                                                       extend_sides_by_eps=self.extend_sides_by_eps)

        return PrivatePolygon(points=new_nodes, convexity=self.convexity)

    # ------------------------------------------------------------------------------------------------------------------
    def _create_polygon_side_extender(self) -> PolygonSideExtender:
        """
        Returns a polygon side extender that extends this polygon.
        """
        return PolygonSideExtender()

# ----------------------------------------------------------------------------------------------------------------------



from abc import ABC
from typing import Any, Dict

from super_scad.scad.ScadWidget import ScadWidget


class ScadSingleChildParent(ScadWidget, ABC):
    """
    Abstract parent widget for SuperSCAD widgets that have a single-child.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, args: Dict[str, Any], child: ScadWidget):
        """
        Object constructor.

        :param child: The child SuperSCAD widget of this single-child parent.
        """
        ScadWidget.__init__(self, args=args)

        self.__child = child
        """
        The child OpenSCAD widget of this single-child parent.
        """

        assert isinstance(child, ScadWidget), f"Child is of type: {child.__class__}"

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def child(self) -> ScadWidget:
        """
        Returns the child of this single-child parent.
        """
        return self.__child

# ----------------------------------------------------------------------------------------------------------------------



import math

from super_scad.scad.Context import Context


class Radius2Sides4n:
    """
    A utility class for converting a radius to sides with a multiple of 4 vertices.
    """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def r2sides(context: Context, radius: float) -> int:
        """
        Replicates the OpenSCAD logic to calculate the number of sides from the radius.

        :param context: The build context.
        :param radius: The radius of the circle.
        """
        if context.fn > 0:
            return max(context.fn, 3)

        return int(math.ceil(max(min(360.0 / context.fa, radius * 2.0 * math.pi / context.fs), 5.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def r2sides4n(context: Context, radius: float) -> int:
        """
        Rounds up the number of sides to a multiple of 4 to ensure points land on all axes.

        :param context: The build context.
        :param radius: The radius of the circle.
        """
        return int(math.floor((Radius2Sides4n.r2sides(context, radius) + 3) / 4) * 4)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateRotate import PrivateRotate
from super_scad.type.Vector2 import Vector2


class Flip2D(ScadSingleChildParent):
    """
    Flips its child about the x or y-axis.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 horizontal: bool | None = None,
                 vertical: bool | None = None,
                 flip_x: bool | None = None,
                 flip_y: bool | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param horizontal: Whether to flip the child widget horizontally (i.e. flip around the y-axis).
        :param vertical: Whether to flip the child widget vertically (i.e. flip around the x-axis).
        :param flip_x: Whether to flip the child widget around the x-asis (i.e. vertical flip).
        :param flip_y: Whether to flip the child widget around the y-asis (i.e. horizontal flip).
        :param child: The child widget to be flipped.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'horizontal', 'vertical'}, {'flip_x', 'flip_y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def horizontal(self) -> bool:
        """
        Returns whether to flip the child widget horizontally (i.e. flip around the y-axis).
        """
        return self._args.get('horizontal', False) or self._args.get('flip_y', False)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vertical(self) -> bool:
        """
        Returns whether to flip the child widget vertically (i.e. flip around the x-axis).
        """
        return self._args.get('vertical', False) or self._args.get('flip_x', False)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def flip_x(self) -> bool:
        """
        Returns whether to flip the child widget around the x-asis (i.e. vertical flip).
        """
        return self.vertical

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def flip_y(self) -> bool:
        """
        Returns whether to flip the child widget around the y-asis (i.e. horizontal flip).
        """
        return self.horizontal

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        angle = Vector2(x=180.0 if self.flip_x else 0.0, y=180.0 if self.flip_y else 0.0)

        return PrivateRotate(angle=angle, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Rotate2D import Rotate2D
from super_scad.transformation.Translate2D import Translate2D
from super_scad.type import Vector2
from super_scad.type.Angle import Angle


class Position2D(ScadSingleChildParent):
    """
    A convenience widget that first rotates its child about the z-axis and then translates its child widget along the
    specified vector.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float | None = None,
                 vector: Vector2 | None = None,
                 x: float | None = None,
                 y: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param angle: The angle of rotation (around the z-axis).
        :param vector: The vector over which the child widget is translated.
        :param x: The distance the child widget is translated to along the x-axis.
        :param y: The distance the child widget is translated to along the y-axis.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'vector'}, {'x', 'y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle(self) -> float:
        """
        Returns the angle of rotation (around the z-axis).
        """
        return Angle.normalize(self._args.get('angle', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector2:
        """
        Returns the vector over which the child widget is translated.
        """
        if 'vector' in self._args:
            return self.uc(self._args['vector'])

        return Vector2(self.x, self.y)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def x(self) -> float:
        """
        Returns distance the child widget is translated to along the x-axis.
        """
        if 'vector' in self._args:
            return self.uc(self._args['vector'].x)

        return self.uc(self._args.get('x', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def y(self) -> float:
        """
        Returns distance the child widget is translated to along the y-axis.
        """
        if 'vector' in self._args:
            return self.uc(self._args['vector'].y)

        return self.uc(self._args.get('y', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        child = self.child

        if self.angle != 0.0:
            child = Rotate2D(angle=self.angle, child=child)

        if self.vector.is_not_origin:
            child = Translate2D(vector=self.vector, child=child)

        return child

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class PrivateProjection(PrivateSingleChildOpenScadCommand):
    """
    Creates 2d drawings from 3d models. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#3D_to_2D_Projection.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, cut: bool, child: ScadWidget) -> None:
        """
        Object constructor.

        :param cut: Whether to cut the 3D model at height 0.0.
        :param child: The child widget.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='projection', args=locals(), child=child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, Dict

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget


class PrivateSingleChildOpenScadCommand(PrivateOpenScadCommand, ScadSingleChildParent):
    """
    Parent widget for OpenSCAD commands with a single child.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, command: str, args: Dict[str, Any], child: ScadWidget):
        """
        Object constructor.

        :param command: The OpenSCAD command.
        :param args: The arguments of the command.
        :param child: The child SuperSCAD widget of this single-child parent.
        """
        PrivateOpenScadCommand.__init__(self, command=command, args=args)
        ScadSingleChildParent.__init__(self, args=args, child=child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import List, Set

from super_scad.d2.PolygonMixin import PolygonMixin
from super_scad.d2.private.PrivatePolygon import PrivatePolygon
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2


class Polygon(PolygonMixin, ScadWidget):
    """
    Widget for creating polygons. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#polygon.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 primary: List[Vector2] | None = None,
                 points: List[Vector2] | None = None,
                 secondary: List[Vector2] | None = None,
                 secondaries: List[List[Vector2]] | None = None,
                 convexity: int | None = None,
                 extend_sides_by_eps: bool | List[bool] | Set[int] | None = None):
        """
        Object constructor.

        :param primary: The list of 2D points of the polygon.
        :param points: Alias for primary.
        :param secondary: The secondary path that will be subtracted from the polygon.
        :param secondaries: The secondary paths that will be subtracted form the polygon.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line
                          through the child widget.
        :param extend_sides_by_eps: Whether to extend sides by eps for a clear overlap.
        """
        ScadWidget.__init__(self, args=locals())
        PolygonMixin.__init__(self, extend_sides_by_eps=extend_sides_by_eps)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'primary'}, {'points'})
        admission.validate_exclusive({'secondary'}, {'secondaries'})
        admission.validate_required({'primary', 'points'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def primary(self) -> List[Vector2]:
        """
        Returns the points of the polygon.
        """
        return self.uc(self._args.get('primary', self._args.get('points')))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def secondaries(self) -> List[List[Vector2]] | None:
        """
        Returns the points of the polygon.
        """
        if 'secondaries' in self._args:
            return [self.uc(point) for point in self._args.get('secondaries')]

        if 'secondary' in self._args:
            return [self.uc(self._args['secondary'])]

        return None

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def nodes(self) -> List[Vector2]:
        """
        Returns the nodes of the polygon.
        """
        return self.primary

    # ------------------------------------------------------------------------------------------------------------------
    def _build_polygon(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        secondaries = self.secondaries
        if secondaries is None:
            return PrivatePolygon(points=self.primary, convexity=self.convexity)

        points = self.primary
        n = 0
        m = n + len(points)
        paths = [list(range(n, m))]
        n = m

        for secondary in secondaries:
            m = n + len(secondary)
            points += secondary
            paths.append(list(range(n, m)))
            n = m

        return PrivatePolygon(points=points, paths=paths, convexity=self.convexity)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Set

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.type.Vector2 import Vector2


class PrivateSquare(PrivateOpenScadCommand):
    """
    Widget for creating squares. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#square.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, size: float | Vector2, center: bool = False):
        """
        Object constructor.

        :param size: The size of the square.
        :param center: Whether the square is centered at the origin.
        """
        PrivateOpenScadCommand.__init__(self, command='square', args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'size'}

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Set

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.scad.ArgumentAdmission import ArgumentAdmission


class PrivateCylinder(PrivateOpenScadCommand):
    """
    Widget for creating cylinders and cones. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#cylinder.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float,
                 radius: float | None = None,
                 diameter: float | None = None,
                 bottom_radius: float | None = None,
                 bottom_diameter: float | None = None,
                 top_radius: float | None = None,
                 top_diameter: float | None = None,
                 center: bool = False,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None):
        """
        Object constructor.

        :param height: The height of the cone or cylinder.
        :param radius: The radius of the cylinder.
        :param diameter: The diameter of the cylinder.
        :param top_radius: The radius at the top of the cone.
        :param top_diameter: The diameter at the top of the cone.
        :param bottom_radius: The radius at the bottom of the cone.
        :param bottom_diameter: The diameter at the bottom of the cone.
        :param center: Whether the cone is centered in the z-direction.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        """
        PrivateOpenScadCommand.__init__(self, command='cylinder', args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'radius'}, {'diameter'})
        admission.validate_exclusive({'top_radius'}, {'top_diameter'}, {'radius'}, {'diameter'})
        admission.validate_exclusive({'bottom_radius'}, {'bottom_diameter'}, {'radius'}, {'diameter'})
        admission.validate_required({'height'},
                                    {'radius', 'diameter', 'bottom_radius', 'bottom_diameter'},
                                    {'radius', 'diameter', 'top_radius', 'top_diameter'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'height':          'h',
                'radius':          'r',
                'diameter':        'd',
                'bottom_radius':   'r1',
                'bottom_diameter': 'd1',
                'top_radius':      'r2',
                'top_diameter':    'd2',
                'fa':              '$fa',
                'fs':              '$fs',
                'fn':              '$fn'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return {'$fa'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'h', 'r', 'r1', 'r2', 'd', 'd1', 'd2', '$fs'}

# ----------------------------------------------------------------------------------------------------------------------



from typing import Tuple

from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateResize import PrivateResize
from super_scad.type.Vector2 import Vector2


class Resize2D(ScadSingleChildParent):
    """
    Modifies the size of the child widget to match the given width and depth. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#resize.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 new_size: Vector2 | None = None,
                 new_width: float | None = None,
                 new_depth: float | None = None,
                 auto: bool | Tuple[bool, bool] | None = None,
                 auto_width: bool | None = None,
                 auto_depth: bool | None = None,
                 convexity: int | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param new_size: The new_size along all two axes.
        :param new_width: The new width (the new size along the x-axis).
        :param new_depth: The new depth (the new size along the y-axis).
        :param auto: Whether to auto-scale any 0-dimensions to match.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        :param child: The child widget to be resized.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'new_size'}, {'new_width', 'new_depth'})
        admission.validate_exclusive({'auto'}, {'auto_width', 'auto_depth'})

        # Handle resizing beyond resolution.

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_size(self) -> Vector2:
        """
        Returns the new_size along all three axes.
        """
        return Vector2(self.new_width, self.new_depth)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_width(self) -> float:
        """
        Returns new width (the new size along the x-axis).
        """
        if 'new_size' in self._args:
            return self.uc(self._args['new_size'].x)

        return self.uc(self._args.get('new_width', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def new_depth(self) -> float:
        """
        Returns the new depth (the new size along the y-axis).
        """
        if 'new_size' in self._args:
            return self.uc(self._args['new_size'].y)

        return self.uc(self._args.get('new_depth', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto(self) -> Tuple[bool, bool]:
        """
        Returns whether to auto-scale the width and depth.
        """
        return self.auto_width, self.auto_depth

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto_width(self) -> bool:
        """
        Returns whether to auto-scale the width (the size along the x-axis).
        """
        if round(self.new_width, 4) == 0.0:  # xxx Use rounding in target units.
            if 'auto' in self._args:
                if isinstance(self._args['auto'], tuple):
                    return self._args['auto'][0]
                else:
                    return self._args['auto']

            if 'auto_width' in self._args:
                return self._args['auto_width']

        return False

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def auto_depth(self) -> bool:
        """
        Returns whether to auto-scale the depth (the size along the y-axis).
        """
        if round(self.new_depth, 4) == 0.0:  # xxx Use rounding in target units.
            if 'auto' in self._args:
                if isinstance(self._args['auto'], tuple):
                    return self._args['auto'][1]
                else:
                    return self._args['auto']

            if 'auto_depth' in self._args:
                return self._args['auto_depth']

        return False

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the convexity.
        """
        return self._args.get('convexity')

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateResize(new_size=self.new_size, auto=self.auto, convexity=self.convexity, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.d3.private.PrivateCube import PrivateCube
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget


class Cube(ScadWidget):
    """
    Widget for creating cubes. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#cube.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: float,
                 center: bool = False):
        """
        Object constructor.

        :param size: The size of the cube.
        :param center: Whether the cube is centered at the origin.
        """
        ScadWidget.__init__(self, args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_required({'size'}, {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cube is centered at the origin.
        """
        return self._args['center']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> float:
        """
        Returns the size of the cube.
        """
        return self.uc(self._args['size'])

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateCube(size=self.size, center=self.center)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Set

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2
from super_scad.type.Vector3 import Vector3


class PrivateMirror(PrivateSingleChildOpenScadCommand):
    """
    Transforms the child widget to a mirror of the original, as if it were the mirror image seen through a plane
    intersecting the origin. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#mirror.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 vector: Vector2 | Vector3,
                 child: ScadWidget):
        """
        Object constructor.

        :param vector:  The normal vector of the origin-intersecting mirror plane used, meaning the vector coming
                        perpendicularly out of the plane. Each coordinate of the original widget is altered such that
                        it becomes equidistant on the other side of this plane from the closest point on the plane.
        :param child: The widget to be mirrored.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='mirror', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'vector': 'v'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'v'}

# ----------------------------------------------------------------------------------------------------------------------



import math

from super_scad.d3.private.PrivateCylinder import PrivateCylinder
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Rotate3D import Rotate3D
from super_scad.transformation.Translate3D import Translate3D
from super_scad.type.Vector3 import Vector3
from super_scad.util.Radius2Sides4n import Radius2Sides4n


class Cylinder(ScadWidget):
    """
    Widget for creating cylinders.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float | None = None,
                 start_point: Vector3 | None = None,
                 end_point: Vector3 | None = None,
                 radius: float | None = None,
                 diameter: float | None = None,
                 center: bool | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 fn4n: bool | None = None):
        """
        Object constructor.

        :param height: The height of the cylinder.
        :param start_point: The start point of the cylinder.
        :param end_point: The end point of the cylinder.
        :param radius: The radius of the cylinder.
        :param diameter: The diameter of the cylinder.
        :param center: Whether the cylinder is centered along the z-as. Defaults to false.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        :param fn4n: Whether to create a cylinder with a multiple of 4 vertices.
        """
        ScadWidget.__init__(self, args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'height'}, {'start_point', 'end_point'})
        admission.validate_exclusive({'radius'}, {'diameter'})
        admission.validate_exclusive({'fn4n'}, {'fa', 'fs', 'fn'})
        admission.validate_required({'height', 'start_point'},
                                    {'height', 'end_point'},
                                    {'radius', 'diameter'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cylinder is centered along the z-as.
        """
        return self._args.get('center', False)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def radius(self) -> float:
        """
        Returns the radius of the cylinder.
        """
        return self.uc(self._args.get('radius', 0.5 * self._args.get('diameter', 0.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def diameter(self) -> float:
        """
        Returns the diameter of the cylinder.
        """
        return self.uc(self._args.get('diameter', 2.0 * self._args.get('radius', 0.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height/length of the cylinder.
        """
        if 'height' in self._args:
            return self.uc(self._args['height'])

        return self.uc((self._args['start_point'] - self._args['end_point']).length)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def start_point(self) -> Vector3:
        """
        Returns the start point of the cylinder.
        """
        if 'start_point' in self._args:
            return self.uc(self._args['start_point'])

        return Vector3(0.0, 0.0, -self.height / 2.0 if self.center else 0.0)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def end_point(self) -> Vector3:
        """
        Returns the end point of the cylinder.
        """
        if 'end_point' in self._args:
            return self.uc(self._args['end_point'])

        return Vector3(0.0, 0.0, self.height / 2.0 if self.center else self.height)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._args.get('fa')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self.uc(self._args.get('fs'))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._args.get('fn')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn4n(self) -> bool | None:
        """
        Returns whether to create a circle with multiple of 4 vertices.
        """
        return self._args.get('fn4n')

    # ------------------------------------------------------------------------------------------------------------------
    def real_fn(self, context: Context) -> int | None:
        """
        Returns the real fixed number of fragments in 360 degrees.
        """
        if self.fn4n:
            return Radius2Sides4n.r2sides4n(context, self.radius)

        return self.fn

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        cylinder = PrivateCylinder(height=self.height,
                                   diameter=self.diameter,
                                   center=self.center,
                                   fa=self.fa,
                                   fs=self.fs,
                                   fn=self.real_fn(context))

        if 'height' in self._args:
            return cylinder

        diff = self.end_point - self.start_point

        return Translate3D(vector=self.start_point,
                           child=Rotate3D(angle_y=math.degrees(math.acos(diff.z / diff.length)),
                                          angle_z=math.degrees(math.atan2(diff.y, diff.x)),
                                          child=cylinder))

# ----------------------------------------------------------------------------------------------------------------------



import math
from typing import List

from super_scad.scad.Context import Context
from super_scad.type import Vector2


class PolygonSideExtender:
    """
    A polygon side extender extends the sides of a polygon by eps for a clear overlap by working at the nodes of the
    polygon.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self):
        """
        Returns a modified list of nodes such that it appears that the sides are extended by eps for a clear overlap.
        """

        self._nodes: List[Vector2] | None = None
        """
        The nodes of the polygon.
        """

        self._inner_angles: List[float] | None = None
        """
        The inner angles of the polygon.
        """

        self._normal_angles: List[float] | None = None
        """
        The absolute angles of the normal of each node.
        """

        self._is_clockwise: bool | None = None
        """
        Whether the nodes are in a clockwise direction.
        """

        self._extend_sides_by_eps: set[int] | None = None
        """
        The set of sides that must be extended by eps for a clear overlap.
        """

        self._current_node: Vector2 | None = None
        """
        The current node of the polygon being processed.
        """

        self._current_inner_angle: float | None = None
        """
        The current inner angle at the current node of the polygon being processed
        """

        self._current_normal_angle: float | None = None
        """
        The current normal angle at the current node of the polygon being processed.
        """

        self._index: int | None = None
        """
        The index of the current node of the polygon being processed.
        """

        self._new_nodes: List[Vector2] = []
        """
        The new nodes of the polygon.
        """

    # ------------------------------------------------------------------------------------------------------------------
    def extend_sides(self,
                     *,
                     context: Context,
                     nodes: List[Vector2],
                     inner_angles: List[float],
                     normal_angles: List[float],
                     is_clockwise: bool,
                     extend_sides_by_eps: set[int]) -> List[Vector2]:
        """
        Returns a modified list of nodes such that it appears that the sides are extended by eps for a clear overlap.

        @param context: The build context.
        @param nodes: The nodes of the polygon.
        @param inner_angles: The inner angles of the polygon.
        @param normal_angles: The absolute angles of the normal of each node.
        @param is_clockwise: Whether the nodes are in a clockwise direction.
        @param extend_sides_by_eps: The set of sides that must be extended by eps for a clear overlap.
        """
        self._nodes = nodes
        self._inner_angles = inner_angles
        self._normal_angles = normal_angles
        self._is_clockwise = is_clockwise
        self._extend_sides_by_eps = extend_sides_by_eps

        self._new_nodes = []
        n = len(self._nodes)
        for self._index in range(n):
            self._set_currents()

            extend_side_by_eps1 = (self._index - 1) % n in self._extend_sides_by_eps
            extend_side_by_eps2 = self._index in self._extend_sides_by_eps

            if self._current_inner_angle <= 180.0:
                # Outer corner.
                if not extend_side_by_eps1 and not extend_side_by_eps2:
                    self._extend_outer_corner_no_sides(context)

                elif not extend_side_by_eps1 and extend_side_by_eps2:
                    self._extend_outer_corner_side2(context)

                elif extend_side_by_eps1 and not extend_side_by_eps2:
                    self._extend_outer_corner_side1(context)

                elif extend_side_by_eps1 and extend_side_by_eps2:
                    self._extend_outer_corner_side1_and_side2(context)

                else:
                    raise ValueError('Someone broke math.')
            else:
                # Inner corner.
                if not extend_side_by_eps1 and not extend_side_by_eps2:
                    self._extend_inner_corner_no_sides(context)

                elif not extend_side_by_eps1 and extend_side_by_eps2:
                    self._extend_inner_corner_side2(context)

                elif extend_side_by_eps1 and not extend_side_by_eps2:
                    self._extend_inner_corner_side1(context)

                elif extend_side_by_eps1 and extend_side_by_eps2:
                    self._extend_inner_corner_side1_and_side2(context)

                else:
                    raise ValueError('Someone broke math.')

        return self._new_nodes

    # ------------------------------------------------------------------------------------------------------------------
    def _set_currents(self) -> None:
        """
        Set the current properties for the current node being processed.
        """
        self._current_node = self._nodes[self._index]
        self._current_inner_angle = self._inner_angles[self._index]
        self._current_normal_angle = self._normal_angles[self._index]

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_inner_corner_no_sides(self, context: Context) -> None:
        """
        Handles the case were at an inner corner neither sides are extended.

        :param context: The build context.
        """
        self._new_nodes.append(self._current_node)

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_inner_corner_side1(self, context: Context) -> None:
        """
        Handles the case were at an inner corner the first side is extended only.

        :param context: The build context.
        """
        if self._is_clockwise:
            angle = self._current_normal_angle + 0.5 * self._current_inner_angle
        else:
            angle = self._current_normal_angle - 0.5 * self._current_inner_angle
        self._new_nodes.append(self._current_node + Vector2.from_polar_coordinates(context.eps, angle))

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_inner_corner_side2(self, context: Context) -> None:
        """
        Handles the case were at an inner corner the second side is extended only.

        :param context: The build context.
        """
        if self._is_clockwise:
            angle = self._current_normal_angle - 0.5 * self._current_inner_angle
        else:
            angle = self._current_normal_angle + 0.5 * self._current_inner_angle
        self._new_nodes.append(self._current_node + Vector2.from_polar_coordinates(context.eps, angle))

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_inner_corner_side1_and_side2(self, context: Context) -> None:
        """
        Handles the case were at an inner corner both the first and the second side must be extended.

        :param context: The build context.
        """
        alpha = 0.5 * (360.0 - self._current_inner_angle)
        length = context.eps / math.sin(math.radians(alpha))
        length = min(length,
                     Vector2.distance(self._current_node, self._nodes[(self._index - 1) % len(self._nodes)]),
                     Vector2.distance(self._current_node, self._nodes[(self._index + 1) % len(self._nodes)]))
        eps0 = Vector2.from_polar_coordinates(length, self._current_normal_angle + 180.0)
        self._new_nodes.append(self._current_node + eps0)

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_outer_corner_no_sides(self, context: Context) -> None:
        """
        Handles the case were at an outer corner neither sides must be extended.

        :param context: The build context.
        """
        self._new_nodes.append(self._current_node)

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_outer_corner_side1(self, context: Context) -> None:
        """
        Handles the case were at an outer corner the first side must be extended only.

        :param context: The build context.
        """
        if self._is_clockwise:
            angle = self._current_normal_angle - 0.5 * self._current_inner_angle - 90.0
        else:
            angle = self._current_normal_angle + 0.5 * self._current_inner_angle + 90.0
        self._new_nodes.append(self._current_node + Vector2.from_polar_coordinates(context.eps, angle))
        self._new_nodes.append(self._current_node)

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_outer_corner_side2(self, context: Context) -> None:
        """
        Handles the case were at an outer corner the second side must be extended only.

        :param context: The build context.
        """
        self._new_nodes.append(self._current_node)
        if self._is_clockwise:
            angle = self._current_normal_angle + 0.5 * self._current_inner_angle + 90.0
        else:
            angle = self._current_normal_angle - 0.5 * self._current_inner_angle - 90.0
        self._new_nodes.append(self._current_node + Vector2.from_polar_coordinates(context.eps, angle))

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_outer_corner_side1_and_side2(self, context: Context) -> None:
        """
        Handles the case were at an outer corner both the first and the second sides must be extended.

        :param context: The build context.
        """
        if self._current_inner_angle >= 90.0:
            self._extend_outer_corner_side1_and_side2_oblique(context)

        else:
            self._extend_outer_corner_side1_and_side2_sharp(context)

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_outer_corner_side1_and_side2_sharp(self, context: Context) -> None:
        """
        Handles the case were at a sharp outer corner both the first and the second sides must be extended.

        :param context: The build context.
        """
        if self._is_clockwise:
            angle1 = self._current_normal_angle - 0.5 * self._current_inner_angle - 90.0
            angle2 = self._current_normal_angle + 0.5 * self._current_inner_angle + 90.0
        else:
            angle1 = self._current_normal_angle + 0.5 * self._current_inner_angle + 90.0
            angle2 = self._current_normal_angle - 0.5 * self._current_inner_angle - 90.0
        self._new_nodes.append(self._current_node + Vector2.from_polar_coordinates(context.eps, angle1))
        self._new_nodes.append(
                self._current_node + Vector2.from_polar_coordinates(context.eps, self._current_normal_angle + 180.0))
        self._new_nodes.append(self._current_node + Vector2.from_polar_coordinates(context.eps, angle2))

    # ------------------------------------------------------------------------------------------------------------------
    def _extend_outer_corner_side1_and_side2_oblique(self, context: Context) -> None:
        """
        Handles the case were at an oblique outer corner both the first and the second sides must be extended.

        :param context: The build context.
        """
        length = context.eps / math.cos(math.radians(0.5 * (180.0 - self._current_inner_angle)))
        self._new_nodes.append(
                self._current_node + Vector2.from_polar_coordinates(length, self._current_normal_angle + 180.0))

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Set, Tuple

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2
from super_scad.type.Vector3 import Vector3


class PrivateResize(PrivateSingleChildOpenScadCommand):
    """
    Modifies the size of the child widget to match the given x and y. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#resize.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 new_size: Vector2 | Vector3 | None = None,
                 auto: bool | Tuple[bool, bool] | Tuple[bool, bool, bool] = None,
                 convexity: int | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param new_size: The new_size along all two axes.
        :param auto: Whether to auto-scale any 0-dimensions to match.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        :param child: The widget to be resized.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='resize', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'new_size': 'newsize'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'newsize'}

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.d3.private.PrivateCylinder import PrivateCylinder
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.util.Radius2Sides4n import Radius2Sides4n


class Cone(ScadWidget):
    """
    Widget for creating cones. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#cylinder.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float,
                 bottom_radius: float | None = None,
                 bottom_diameter: float | None = None,
                 top_radius: float | None = None,
                 top_diameter: float | None = None,
                 center: bool = False,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 fn4n: bool | None = None):
        """
        Object constructor.

        :param height: The height of the cone.
        :param bottom_radius: The radius at the bottom of the cone.
        :param bottom_diameter: The diameter at the bottom of the cone.
        :param top_radius: The radius at the top of the cone.
        :param top_diameter: The diameter at the top of the cone.
        :param center: Whether the cone is centered in the z-direction.
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        :param fn4n: Whether to create a cone with a multiple of 4 vertices.
        """
        ScadWidget.__init__(self, args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'bottom_radius'}, {'bottom_diameter'})
        admission.validate_exclusive({'top_radius'}, {'top_diameter'})
        admission.validate_exclusive({'fn4n'}, {'fa', 'fs', 'fn'})
        admission.validate_required({'height'},
                                    {'bottom_radius', 'bottom_diameter'},
                                    {'top_radius', 'top_diameter'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cone is centered along the z-as.
        """
        return self._args['center']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def bottom_radius(self) -> float:
        """
        Returns the bottom radius of the cone.
        """
        return self.uc(self._args.get('bottom_radius', 0.5 * self._args.get('bottom_diameter', 0.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def bottom_diameter(self) -> float:
        """
        Returns the bottom diameter of the cone.
        """
        return self.uc(self._args.get('bottom_diameter', 2.0 * self._args.get('bottom_radius', 0.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def top_radius(self) -> float:
        """
        Returns the top radius of the cone.
        """
        return self.uc(self._args.get('top_radius', 0.5 * self._args.get('top_diameter', 0.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def top_diameter(self) -> float:
        """
        Returns the top diameter of the cone.
        """
        return self.uc(self._args.get('top_diameter', 2.0 * self._args.get('top_radius', 0.0)))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height of the cone.
        """
        return self.uc(self._args.get('height', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._args.get('fa')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self.uc(self._args.get('fs'))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._args.get('fn')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn4n(self) -> bool | None:
        """
        Returns whether to create a circle with multiple of 4 vertices.
        """
        return self._args.get('fn4n')

    # ------------------------------------------------------------------------------------------------------------------
    def real_fn(self, context: Context) -> int | None:
        """
        Returns the real fixed number of fragments in 360 degrees.
        """
        if self.fn4n:
            return Radius2Sides4n.r2sides4n(context, max(self.bottom_radius, self.top_radius))

        return self.fn

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateCylinder(height=self.height,
                               bottom_diameter=self.bottom_diameter,
                               top_diameter=self.top_diameter,
                               center=self.center,
                               fa=self.fa,
                               fs=self.fs,
                               fn=self.real_fn(context))

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.d3.private.PrivateCube import PrivateCube
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector3 import Vector3


class Cuboid(ScadWidget):
    """
    Class for cuboids.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: Vector3 | None = None,
                 width: float | None = None,
                 depth: float | None = None,
                 height: float | None = None,
                 center: bool = False):
        """
        Object constructor.

        :param size: The size of the cuboid.
        :param width: The width (the size along the x-axis) of the cuboid.
        :param depth: The depth (the size along the y-axis) of the cuboid.
        :param height: The height (the size along the y-axis) of the cuboid.
        :param center: Whether the cuboid is centered at the origin.
        """
        ScadWidget.__init__(self, args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'size'}, {'width', 'depth', 'height'})
        admission.validate_required({'size', 'width'},
                                    {'size', 'depth'},
                                    {'size', 'height'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the cuboid is centered at the origin.
        """
        return self._args['center']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> Vector3:
        """
        Returns the size of the cuboid.
        """
        return Vector3(x=self.width, y=self.depth, z=self.height)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def width(self) -> float:
        """
        Returns the width of the cuboid.
        """
        if 'size' in self._args:
            return self.uc(self._args['size'].x)

        return self.uc(self._args['width'])

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def depth(self) -> float:
        """
        Returns the depth of the cuboid.
        """
        if 'size' in self._args:
            return self.uc(self._args['size'].y)

        return self.uc(self._args['depth'])

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height of the cuboid.
        """
        if 'size' in self._args:
            return self.uc(self._args['size'].z)

        return self.uc(self._args['height'])

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateCube(size=self.size, center=self.center)

# ----------------------------------------------------------------------------------------------------------------------
